/*  serial port interface for Palm Portable Keyboard
 *  Copyright (C) 2000  
 *  Dan Morris <dmorris@cs.brown.edu> (initial and windows version)
 *  Ralf Ackermann <rac@KOM.tu-darmstadt.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *  Add ksim.c (2003.2.16) Thanks ...
 */

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <string.h>
#include <termio.h>

#include <asm/sharp_char.h>
#include <asm/sharp_keycode.h>
#include <linux/kbd_ll.h>


/*
protocol is:
 - without modifier:
   key press => key code
   key release => 2 * (key code & (1 << 7)) (most significant bit set)
 - with modifier:
   key press => key code
   key release => 1 * (key code & (1 << 7)) (most significant bit set)
*/

/*
implementation:
 - state machine with states:
   shift
   ctrl
   alt
*/

#define STATE_PRESSED 0
#define STATE_RELEASED 1

#define FUNCKEY 0x22

#define RETOUR_OK	0
#define RETOUR_ERROR	-1

typedef struct key_mapping
{
    char *name1;
    char *name2;
    long normal;
    long shift;
    long ctrl;
    long alt;
}
key_mappings;

/* prototypes */

int read_key_mapping(char *map_file);
int init_ser(char *ser_device);
int set_ser_param(int fd);
void flush_ser(void);
int shutdown_ser(void);
unsigned char get_1_char(void);

static void onekey(int c, FILE *fp);
static void onekey1(int c, FILE *fp, int onoff);
static int onekey_s( unsigned long c, int fd );

/* (fd == -1) shows that fd is not valid */
static int fd=-1;
static key_mappings keymap[256];
static FILE *a2x_pipe=NULL;

int read_key_mapping(char *map_file)
{
    FILE *in_file;
    char line[256];

    int  i;
    int  id;
    char name1[256];
    char name2[256];
    long normal,shift,ctrl,alt;

    for (i=0; i < 256; i++)
    {
        keymap[i].name1  = NULL;
        keymap[i].name2  = NULL;
	keymap[i].normal = 0;
	keymap[i].shift  = 0;
	keymap[i].ctrl   = 0;
	keymap[i].alt    = 0;
    }

    in_file=fopen(map_file, "r");

    if (in_file != NULL)
    {
        while (!feof(in_file))
        {
            if (NULL != fgets(line, 256, in_file))
            {
		normal = shift = ctrl = alt = 0;
                sscanf(line, "%x %s %s %lx %lx %lx %lx", &id, name1, name2, &normal, &shift, &ctrl, &alt);
                keymap[id].name1=strdup(name1);
                keymap[id].name2=strdup(name2);
                keymap[id].normal = normal;
                keymap[id].shift = shift;
		keymap[id].ctrl = ctrl;
		keymap[id].alt = alt;
            }
        }
        fclose(in_file);
        return(RETOUR_OK);
    }
    else {
        return(RETOUR_ERROR);
    }
}

int init_ser(char *ser_device)
{
    /* fd is static to this module but not to be seen outside */
    if ((fd=open(ser_device, O_RDWR | O_NOCTTY | O_NONBLOCK, 0)) == -1) {
        fprintf(stderr, "Error opening serial device !\n");
        return (RETOUR_ERROR);
    }
    else {
        if (set_ser_param(fd) == RETOUR_ERROR) {
            printf("Couldn't set serial parameters !\n");
            close(fd);
            fd=-1;
            return (RETOUR_ERROR);
        }
        else {
            /* printf("Sucessfully opened serial device !\n"); */
            return (RETOUR_OK);
        }
    }
}

int set_ser_param(int fd)
{
    int line;
    struct termio sgtty;

    if (ioctl(fd, TCGETA, &sgtty) < 0) {
        fprintf(stderr, "can't TCGETA\n");
        fgetc(stdin);
        return (-1);
    }
    sgtty.c_iflag=(IMAXBEL | IGNBRK);
    sgtty.c_oflag=0;
    sgtty.c_lflag=0;
    sgtty.c_cflag=(CLOCAL | HUPCL | B9600 | CS8 | CREAD);
    sgtty.c_cflag &=~(PARENB | PARODD);
    sgtty.c_cc[VTIME]='\0';
    sgtty.c_cc[VMIN]='\0';
    if (ioctl(fd, TCSETA, &sgtty) < 0) {
        fprintf(stderr, "can't TCSETA\n");
        fgetc(stdin);
        return (RETOUR_ERROR);
    }

    ioctl (fd, TIOCMGET, &line);
    line |= TIOCM_DTR;
    ioctl (fd, TIOCMSET, &line);
    printf("DTR set to HIGH\n");

    sleep(1);

    ioctl (fd, TIOCMGET, &line);
    line &= ~TIOCM_RTS;
    ioctl (fd, TIOCMSET, &line);

    printf("RTS set to LOW\n");
    usleep(1000);

    ioctl (fd, TIOCMGET, &line);
    line |= TIOCM_RTS;
    ioctl (fd, TIOCMSET, &line);
    printf("RTS set to HIGH\n");

    sleep(1);

    flush_ser();

    return (RETOUR_OK);
}

void flush_ser(void)
{
    ioctl(fd, TCFLSH, 0);
}

int shutdown_ser(void)
{
    if (close(fd) == -1) {
        fd=-1;
        return (RETOUR_ERROR);
    }
    else {
        fd=-1;
        return (RETOUR_OK);
    }
}

unsigned char get_1_char(void)
{
    unsigned char recv_char;
    int r_code;

    do {
        /* wait for a char to arrive */
        r_code = read(fd, &recv_char, 1);
    }
    while (r_code != 1);

    return (recv_char);
}


#if 0
void main(void)
{
	int	i;

	read_key_mapping("ser_code_ksim.ini");

	for ( i = 0; i < 255; i++ ) {
		printf( "%d : (%s), (%s), (%lx), (%lx), (%lx)\n", i, 
			keymap[i].name1, 
			keymap[i].name2,
			keymap[i].normal,
			keymap[i].shift,
			keymap[i].ctrl,
			keymap[i].alt
		 );
	}
}
#else
int main(int argc, char **argv)
{
    unsigned char buffer;
    int fd;
    int func_on=0;
    int incode;
    int onoff;
    int state;
    char state_char;
    int keycode;
    int shift_on = 0;	/* shift flag */
    int ctrl_on = 0;	/* ctrl flag */
    long code;
    char *name;
    int i, m = 0;
    FILE* p;
    int fds;

    if ( (fds = open("/dev/sharp_kbdctl", O_RDWR)) < 0 ) {
	return 1;
    }
/*
    if ( (p = fopen("/dev/kbdsim", "w")) == NULL )
	return -1;
*/

    read_key_mapping("/usr/local/bin/ser_code_ksim.ini");
    fd=init_ser("/dev/ttyS0");

    while (1)
    {
        buffer = get_1_char();
	if (buffer & (1 << 7))
	{
		state=STATE_RELEASED;
		state_char='R';
		onoff = 0;
	} else {
		state=STATE_PRESSED;
		state_char='P';
		onoff = 1;
	}

	keycode = buffer & 0x7f;

	/* shift flag */
	if ( (keycode == 0x58) || (keycode == 0x59) ) {
		if ( onoff ) {
			shift_on = 1;
		}else{
			shift_on = 0;
		}
		continue;
	}

	/* ctrl flag */
	if ( keycode == 0x1a ) {
		if ( onoff ) {
			ctrl_on = 1;
		}else{
			ctrl_on = 0;
		}
		continue;		/* control key no baaidake wakeru */
	}

	if ( ctrl_on && (keymap[keycode].ctrl != 0) ) {
		code = keymap[keycode].ctrl;
		name = "ctrl-code";
	}else if ( shift_on && (keymap[keycode].shift != 0) ) {
		code = keymap[keycode].shift;
		name = keymap[keycode].name2;
	}else{
		code = keymap[keycode].normal;
		name = keymap[keycode].name1;
	}

        /*
		printf("%02x %2x (%c) => %s (%d)\n", (unsigned char)buffer, keycode, state_char, name, code );
		fflush(stdout);
	*/

#if 0
	if ( code > 0xff ) {
		if ( onoff ) {
			if ( code > 0xffff ) {
				onekey1( (code >> 24) & 0xff, p, 1 );
				onekey1( (code >> 16) & 0xff, p, 1 );
				onekey1( (code) & 0xff, p, 1 );

				onekey1( (code) & 0xff, p, 0 );
				onekey1( (code >> 16) & 0xff, p, 0 );
				onekey1( (code >> 24) & 0xff, p, 0 );
			}else{
				onekey1( (code >> 16) & 0xff, p, 1 );
				onekey1( (code) & 0xff, p, 1 );

				onekey1( (code) & 0xff, p, 0 );
				onekey1( (code >> 16) & 0xff, p, 0 );
			}
		}
	}else{
		onekey1( code, p, onoff );
	}
#else
	if ( onoff ) {
		onekey_s( code, fds );
	}
#endif

    }

/*
	fclose( p );
*/
	close( fds );
}
#endif

static void onekey(int c, FILE *fp) {
	unsigned char buf[2];

	buf[0] = c; buf[1] = 1;
	fwrite(buf, 2, 1, fp);
	fflush(fp);

	buf[0] = c + 0x80; buf[1] = 0;
	fwrite(buf, 2, 1, fp);
	fflush(fp);
}

/* onoff is 0 or 1 */
static void onekey1(int c, FILE *fp, int onoff) {
	unsigned char buf[2];

	buf[0] = c; buf[1] = onoff;
	fwrite(buf, 2, 1, fp);
	fflush(fp);
}

static int onekey_s( unsigned long c, int fd ) {
	int	rc;

	rc = ioctl( fd, SHARP_KBDCTL_SENDKEY, c );
	if ( rc ) {
		printf( "ioctl error %d\n", rc );
		return 1;
	}

	return 0;
}
